local addon = LFGTimer;

if not addon then
	return;
end

local lsm = LibStub("LibSharedMedia-3.0");
local L = LibStub("AceLocale-3.0"):GetLocale("LFGTimer");

local Recycler; -- table recycler for the frame update functions. should help minimize the rampant memory usage growth
do
	local _Acquire = function(self, ...)
		local t = next(self.FreeTables);
		if not t then
			t = {};
		else
			self.FreeTables[t] = nil;
		end
		for i = 1, select("#", ...) do
			t[i] = select(i, ...);
		end
		return t;
	end
	local _Release = function(self, t, emptyChildren)
		if self.FreeTables[t] then -- table already released
			return;
		end
		if emptyChildren then
			for k, v in pairs(t) do
				t[k] = nil;
				if type(k) == "table" and t ~= k then -- if the key is a table, then recycle it but avoid following circular references
					self:Release(k);
				end
				if type(v) == "table" and t ~= v then
					self:Release(v);
				end
			end
		else
			wipe(t);
		end
		self.FreeTables[t] = true;
	end
	Recycler = setmetatable({}, {
		__index = {
			Acquire = _Acquire,
			Release = _Release,
		},
	});
	Recycler.FreeTables = {};
end

local TOTAL_LFG_PARTY_SIZE = 5; -- 'global' used to determine the size of the group

local db; -- set in :SetupFloatingFrame

function addon:UpdateFrameDB(database)
	db = database;
end

function addon:SetupFloatingFrame()
	db = self.db.profile;
	self:UpdateFrameShown(); -- create and show the proper frame
end

function addon:CreateCloneFrame()
	if self.Clone then
		self.Clone:Show();
		return;
	end
	self.Clone = CreateFrame("Frame", "LFGTimerCloneFrame", UIParent);
	local frame = self.Clone;
	local name = frame:GetName();
	frame:SetWidth(350);
	frame:SetHeight(153);
	frame:SetFrameStrata("TOOLTIP");
	frame:SetMovable(true);
	frame:SetClampedToScreen(true);
	frame:SetUserPlaced(true);
	self:SetAnchor(self.Clone); -- put the frame where it should be
	frame:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		tile = true,
		edgeSize = 16,
		tileSize = 16,
		insets = {
			left = 5,
			right = 5,
			top = 5,
			bottom = 5,
		},
	});
	
	frame.Drag = CreateFrame("Button", frame:GetName() .. "DragButton", frame);
	frame.Drag:SetPoint("TOPLEFT", frame, "TOPLEFT", 10, -5);
	frame.Drag:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -10, -5);
	frame.Drag:SetHeight(8);
	frame.Drag:SetHighlightTexture("Interface\\FriendsFrame\\UI-FriendsFrame-HighlightBar");
	frame.Drag:SetScript("OnMouseDown", function()
		frame:StartMoving()
	end);
	frame.Drag:SetScript("OnMouseUp", function()
		frame:StopMovingOrSizing();
		addon:UpdatePosition(addon.Clone);
	end);
	
	frame.Title = frame:CreateFontString(name .. "Title", "ARTWORK", "GameFontHighlight");
	frame.Title:SetText(ASSEMBLING_GROUP);
	frame.Title:SetPoint("TOP", frame, "TOP", 0, -15);
	
	frame.Statistic = frame:CreateFontString(name .. "Statistic", "ARTWORK", "GameFontNormal");
	frame.Statistic:SetHeight(25);
	frame.Statistic:SetWidth(330);
	frame.Statistic:SetJustifyH("CENTER");
	frame.Statistic:SetJustifyV("MIDDLE");
	frame.Statistic:SetPoint("TOP", frame, "TOP", 0, -90);
	
	frame.WaitTime = frame:CreateFontString(name .. "WaitTime", "ARTWORK", "GameFontNormal");
	frame.WaitTime:ClearAllPoints();
	frame.WaitTime:SetPoint("BOTTOM", frame, "BOTTOM", 0, 38);
	frame.WaitTime:Hide();
	
	frame.LookingFor = frame:CreateFontString(name .. "LookingFor", "ARTWORK", "GameFontHighlight");
	frame.LookingFor:SetText(SEARCHING_FOR_GROUPS_NEEDS);
	frame.LookingFor:SetJustifyH("CENTER");
	frame.LookingFor:SetPoint("BOTTOM", frame, "BOTTOM", 0, 14);
	
	local last;
	
	for i = 1, 3 do
		local tex = frame:CreateTexture(name .. "RoleIcon" .. i, "ARTWORK", "LFDSearchStatusRoleIconTemplate");
		tex:ClearAllPoints();
		if i == 1 then -- first role icon
			tex:SetPoint("LEFT", frame.LookingFor, "RIGHT", 2, 0);
		else
			tex:SetPoint("LEFT", last, "RIGHT", 2, 0);
		end
		last = tex; --]]
		frame["RoleIcon" .. i] = tex;
	end
	
	frame.Tank1 = CreateFrame("Frame", name .. "Tank1", frame, "LFDSearchStatusPlayerTemplate");
	frame.Tank1:SetPoint("TOPLEFT", frame, "TOPLEFT", 25, -40);
	frame.Tank1.texture:SetTexCoord(GetTexCoordsForRole("TANK"));
	
	frame.Healer1 = CreateFrame("Frame", name .. "Healer1", frame, "LFDSearchStatusPlayerTemplate");
	frame.Healer1:SetPoint("LEFT", frame.Tank1, "RIGHT", 5, 0);
	frame.Healer1.texture:SetTexCoord(GetTexCoordsForRole("HEALER"));
	
	last = frame.Healer1;
	
	for i = 1, 3 do
		local f = CreateFrame("Frame", name .. "Damage" .. i, frame, "LFDSearchStatusPlayerTemplate");
		f:SetPoint("LEFT", last, "RIGHT", 5, 0);
		f.texture:SetTexCoord(GetTexCoordsForRole("DAMAGER"));
		frame["Damage" .. i] = f;
		last = f;
	end
	
	last = nil;
	
	self:CloneUpdateRoles();
	
	frame:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b);
	frame:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b);
	frame:SetScript("OnUpdate", self.StringHooks["CloneUpdate"]);
end

function addon:CloneUpdate()
	local frame = self.Clone;
	local lfgData = Recycler:Acquire(GetLFGQueueStats());
	local lfgMode = Recycler:Acquire(GetLFGMode());
--[[	local lfgData = {GetLFGQueueStats()};
	local lfgMode = {GetLFGMode()}; --]]
	frame:SetHeight(153); -- reset frame height
	
	self:CloneUpdateRoles();
	
	if (not lfgMode[1] or lfgMode[1] == "abandonedInDungeon") and not lfgData[1] then
		frame.Statistic:SetText(L["You are not in an LFG queue"]);
		frame.WaitTime:Hide();
	elseif lfgMode[1] == "proposal" then
		frame.Statistic:SetText(L["Found a group"]);
		if lfgMode[2] == "accepted" then
			frame.Statistic:SetText(L["Found a group"] .. "\n" .. L["Waiting on group members"]);
			frame:SetHeight(170); -- set for 3 lines of text
		end
	elseif(lfgMode[1] == "queued" or lfgMode[1] == "listed" or lfgMode[1] == "rolecheck") and not lfgData[1] then -- queued, waiting on data
		if lfgData[13] then
			local t = GetTime() - lfgData[13];
			frame.WaitTime:SetFormattedText(TIME_IN_QUEUE, (t >= 60) and SecondsToTime(t) or LESS_THAN_ONE_MINUTE);
			frame.WaitTime:Show();
			frame:SetHeight(158);
		end
		frame.Statistic:SetText(L["Acquiring LFG queue data..."]);
	elseif lfgMode[1] == "lfgparty" then -- in an instance
		frame.Statistic:SetText(L["You are in an LFG party"]);
	else
		local elapsedTime = GetTime() - lfgData[13];
		frame.Statistic:SetFormattedText(LFG_STATISTIC_AVERAGE_WAIT, lfgData[12] == -1 and TIME_UNKNOWN or SecondsToTime(lfgData[12], false, false, 1));
		frame.WaitTime:SetFormattedText(TIME_IN_QUEUE, (elapsedTime >= 60) and SecondsToTime(elapsedTime) or LESS_THAN_ONE_MINUTE);
		frame.WaitTime:Show();
		frame:SetHeight(158);
--		frame.Statistic:SetFormattedText(LFG_STATISTIC_AVERAGE_WAIT .. "\n" .. L["Time in queue: %s"], lfgData[12] == -1 and TIME_UNKNOWN or SecondsToTime(lfgData[12], false, false, 1), SecondsToTime(GetTime() - db.LFGStartTime, false, false, 1));
	end
	
	if lfgData[3] == nil then
		lfgData[3] = 1;
	end
	if lfgData[4] == nil then
		lfgData[4] = 1;
	end
	if lfgData[5] == nil then
		lfgData[5] = 3;
	end
	
	self:CloneUpdateFound({
		tank = (1 - lfgData[3]),
		healer = (1 - lfgData[4]),
		dps = (3 - lfgData[5]),
	});
	Recycler:Release(lfgData);
	Recycler:Release(lfgMode);
end

function addon:CloneUpdateFound(t)
--[[
	t: # of group members found
	t.tank: # of tanks
	t.healer: # of healers
	t.dps: # of damage dealers
--]]
	local frame = self.Clone;
	if t.tank > 0 then -- have a tank
		SetDesaturation(frame.Tank1.texture, false);
		frame.Tank1.cover:Hide();
	else
		SetDesaturation(frame.Tank1.texture, true);
		frame.Tank1.cover:Show();
	end
	if t.healer > 0 then -- have a tank
		SetDesaturation(frame.Healer1.texture, false);
		frame.Healer1.cover:Hide();
	else
		SetDesaturation(frame.Healer1.texture, true);
		frame.Healer1.cover:Show();
	end
	for i = 1, 3 do
		local f = frame["Damage" .. i];
		if t.dps >= i then
			SetDesaturation(f.texture, false);
			f.cover:Hide();
		else
			SetDesaturation(f.texture, true);
			f.cover:Show();
		end
	end
end

function addon:CloneUpdateRoles()
	local frame = self.Clone;
	local roles = Recycler:Acquire(GetLFGRoles());
--	local roles = {GetLFGRoles()};
	local i = 1;
	local icon;
	if roles[2] then
		icon = frame["RoleIcon" .. i];
		icon:SetTexCoord(GetTexCoordsForRole("TANK"));
		icon:Show();
		i = i + 1;
	end
	if roles[3] then
		icon = frame["RoleIcon" .. i];
		icon:SetTexCoord(GetTexCoordsForRole("HEALER"));
		icon:Show();
		i = i + 1;
	end
	
	if roles[4] then
		icon = frame["RoleIcon" .. i];
		icon:SetTexCoord(GetTexCoordsForRole("DAMAGER"));
		icon:Show();
		i = i + 1;
	end
	local eW = 27 * (i - 1); -- extra width
	while i <= 3 do
		icon = frame["RoleIcon" .. i];
		icon:Hide();
		i = i + 1;
	end
	frame.LookingFor:SetPoint("BOTTOM", -eW / 2, 14);
	Recycler:Release(roles);
end

function addon:CreateCustomFrame()
	if addon.Frame then -- already created the frame
		addon.Frame:Show(); -- so show it
		return
	end
	self.Frame = CreateFrame("Frame", "LFGTimerCustomFrame", UIParent);
	self.Frame.Lines = {};
	self.TextFrames = {};
	local frame = addon.Frame;
	
	frame:SetMovable(true);
	frame:SetClampedToScreen(true);
	frame:SetScript("OnUpdate", function()
		addon:UpdateFrameData();
	end);
	frame:SetUserPlaced(true);
	frame:SetWidth(150);
	frame:SetHeight(150);
	self:SetAnchor(self.Frame); -- put the frame where it should be
	
	frame.Drag = CreateFrame("Button", frame:GetName() .. "DragButton", frame);
	frame.Drag:SetPoint("TOPLEFT", frame, "TOPLEFT", 10, -5);
	frame.Drag:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -10, -5);
	frame.Drag:SetHeight(8);
	frame.Drag:SetHighlightTexture("Interface\\FriendsFrame\\UI-FriendsFrame-HighlightBar");
	frame.Drag:SetScript("OnMouseDown", function()
		frame:StartMoving()
	end);
	frame.Drag:SetScript("OnMouseUp", function()
		frame:StopMovingOrSizing();
		addon:UpdatePosition(addon.Frame);
	end);
	
	local inset = addon:CalcInset();
	
	frame.Container = CreateFrame("Frame", frame:GetName() .. "TextContainer", frame);
	
	frame.Container:SetPoint("TOP", frame.Drag, "BOTTOM", 0, -5);
	frame.Container:SetPoint("BOTTOM", frame, "BOTTOM", 0, inset);
	frame.Container:SetPoint("LEFT", frame, "LEFT", inset, 0);
	frame.Container:SetPoint("RIGHT", frame, "RIGHT", -inset, 0);
	
	self:UpdateFrameData();
	self:UpdateFrameTextures();
	self:UpdateFrameFonts();
end

function addon:ShowFrame() -- returns if the frame should be shown
	local lfgData = Recycler:Acquire(GetLFGQueueStats());
	local lfgMode = Recycler:Acquire(GetLFGMode());
	local ret = false;
--[[	local lfgData = {GetLFGQueueStats()}; -- possibly cache these values and update on an event (LFG_UPDATE maybe)
	local lfgMode = {GetLFGMode()}; --]]
	if not db.Frame.Show then
		ret = false;
	elseif (not lfgMode[1] or lfgMode[1] == "abandonedInDungeon") and not lfgData[1] then -- not queued
		ret = db.Frame.ShowWhen.Not;
	elseif lfgMode[1] == "proposal" then -- queue popped, waiting to join
		ret = db.Frame.ShowWhen.Found;
	elseif (lfgMode[1] == "queued" or lfgMode[1] == "listed" or lfgMode[1] == "rolecheck") and not lfgData[1] then -- queued, waiting on data
		ret = db.Frame.ShowWhen.Acquire;
	elseif lfgMode[1] == "lfgparty" then -- in an instance
		ret = db.Frame.ShowWhen.Party;
	elseif lfgMode[1] == "queued" and lfgData[1] then -- queued, have data
		ret = db.Frame.ShowWhen.Wait;
	else
		ret = false; -- wtf is going on with LFG, hide the frame
	end
	Recycler:Release(lfgData);
	Recycler:Release(lfgMode);
	return ret;
end

function addon:UpdateFrameShown()
	if self.Clone then
		self.Clone:Hide();
	end
	if self.Frame then
		self.Frame:Hide();
	end
	if self:ShowFrame() then
		if db.Frame.Option == "clone" then
			if not self.Clone then
				addon:CreateCloneFrame();
			end
			self.Clone:Show();
		elseif db.Frame.Option == "custom" then
			if not self.Frame then
				addon:CreateCustomFrame();
			end
			self.Frame:Show();
		end
	end
end

function addon:CalcInset()
	if db.Frame.Border.Texture == "None" then
		return 0;
	elseif db.Frame.Border.Texture == "Blizzard Chat Bubble" then -- extra wide border texture
		return db.Frame.Border.EdgeSize;
	else
		return math.floor(db.Frame.Border.EdgeSize / 2);
	end
end

function addon:ClearLines() -- hide all lines of text and empty the list of lines on the frame
	if not self.Frame then -- no custom frame. why was this called?
		return;
	end
	Recycler:Release(self.Frame.Lines);
	self.Frame.Lines = Recycler:Acquire();
	for i, v in ipairs(self.TextFrames) do
		v:Hide();
		v.Text:SetText("Empty");
	end
end

function addon:GetTextFrame(i)
	if not self.TextFrames[i] then
		local font = lsm:Fetch("font", db.Frame.Text.Font);
		local size = db.Frame.Text.Size;
		local flags = db.Frame.Text.Outline;
		local color = db.Frame.Text.Color;
		local frame = CreateFrame("Frame", "LFGTimerTextFrame" .. i, self.Frame);
		frame.Text = frame:CreateFontString(frame:GetName() .. "Text", nil, "GameFontNormalSmall"); -- starter font object
		frame.Text:SetJustifyH("LEFT");
		frame.Text:SetJustifyV("TOP");
		frame.Text:SetText("Initialized");
		frame.Text:SetFont(font, size, flags);
		frame.Text:SetTextColor(color.r, color.g, color.b, color.a);
		frame.Text:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5); -- bind the text to the top left corner of the frame. will be "broken" for right-to-left locales
		frame.Text:Show(); -- make sure the text is shown
		self.TextFrames[i] = frame;
	end
	return self.TextFrames[i];
end

function addon:AddLine(text) -- add a line of text to the custom frame
	if not self.Frame then -- why are we adding lines to a frame that does not exist?
		return;
	end
	local frame = self:GetTextFrame(#self.Frame.Lines + 1);
	frame.Text:SetText(text);
	table.insert(self.Frame.Lines, frame);
end

function addon:UpdateSize()
	local inset = self:CalcInset();
	local last = nil;
	local maxWidth, height = 0, 25;
	for i, v in pairs(self.Frame.Lines) do
		v:ClearAllPoints();
		if not last then -- first line
			v:SetPoint("TOP", self.Frame.Container, "TOP", 0, inset / 3);
		else
			v:SetPoint("TOP", last, "BOTTOM", 0, 0);
		end
		v:SetPoint("LEFT", self.Frame.Container, "LEFT", (-2 * inset / 3), 0);
		v:SetWidth(v.Text:GetWidth() + (inset / 3));
		v:SetHeight(v.Text:GetHeight());
		height = height + v:GetHeight();
		if v:GetWidth() > maxWidth then
			maxWidth = v:GetWidth();
		end
		v:Show();
		last = v;
	end
	self.Frame:SetWidth(maxWidth + 18); -- add offset to allow for extra room in the frame
	self.Frame:SetHeight(height);
end

function addon:UpdatePosition(frame)
	local anchor = db.Frame.Position;
	local x, y = 0, 0;
	local c1, c2 = {UIParent:GetCenter()}, {frame:GetCenter()};
	x = c2[1] - c1[1];
	y = c2[2] - c1[2];
	anchor.AnchorAtSelf = "CENTER";
	anchor.AnchoredTo = "UIParent";
	anchor.AnchorAtOther = "CENTER";
	anchor.AnchorXOffset = x;
	anchor.AnchorYOffset = y;
	self:SetAnchor(frame);
end

function addon:SetAnchor(frame)
	local anchor = db.Frame.Position;
	frame:ClearAllPoints();
	frame:SetPoint(anchor.AnchorAtSelf, anchor.AnchoredTo, anchor.AnchorAtOther, anchor.AnchorXOffset, anchor.AnchorYOffset);
end

function addon:UpdateFrameData()
	self:ClearLines();
	local lfgData = Recycler:Acquire(GetLFGQueueStats());
	local lfgMode = Recycler:Acquire(GetLFGMode());
	if (not lfgMode[1] or lfgMode[1] == "abandonedInDungeon") and not lfgData[1] then -- not queued
		self:AddLine(L["You are not in an LFG queue"]);
	elseif lfgMode[1] == "proposal" then -- queue popped, waiting to join
		self:AddLine(L["Found a group"]);
		if lfgMode[2] == "accepted" then
			self:AddLine(L["Waiting on group members"]);
		end
	elseif (lfgMode[1] == "queued" or lfgMode[1] == "listed" or lfgMode[1] == "rolecheck") and not lfgData[1] then -- queued, waiting on data
		self:AddLine(L["Acquiring LFG queue data..."]);
		if lfgData[13] then
			self:AddLine((L["Time in queue: %s"]):format(SecondsToTime(GetTime() - lfgData[13])));
		end
	elseif lfgMode[1] == "lfgparty" then -- in an instance
		self:AddLine(L["You are in an LFG party"]);
	elseif lfgMode[1] == "queued" and lfgData[1] then -- queued, have data
		if db.Frame.ShowTank then
			self:AddLine((L["%s queue times: %s"]):format(L["Tank"], lfgData[9] == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData[9])));
		end
		if db.Frame.ShowHealer then
			self:AddLine((L["%s queue times: %s"]):format(L["Healer"], lfgData[10] == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData[10])));
		end
		if db.Frame.ShowDPS then
			self:AddLine((L["%s queue times: %s"]):format(L["DPS"], lfgData[11] == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData[11])));
		end
		if db.Frame.ShowSelf then
			self:AddLine((L["Estimated queue time: %s"]):format(lfgData[12] == -1 and L["Instant or Unknown"] or SecondsToTime(lfgData[12])));
		end
		self:AddLine((L["Time in queue: %s"]):format(SecondsToTime(GetTime() - lfgData[13])));
		if db.Frame.ShowParty then
			self:AddLine((L["Party members found: %d/%d"]):format(TOTAL_LFG_PARTY_SIZE - lfgData[3] - lfgData[4] - lfgData[5], TOTAL_LFG_PARTY_SIZE));
		end
	else
		self:AddLine(L["Found an odd state, please submit the data that the addon prints to the chat window"]);
		self:TableDump({GetLFGMode = lfgMode, GetLFGQueueStats = lfgData}); -- ask people to submit this, i'm not able to produce it and i believe that some other states should be handled
	end
	self:UpdateSize();
	Recycler:Release(lfgData);
	Recycler:Release(lfgMode);
end

function addon:UpdateFrameTextures()
	local inset = self:CalcInset();
	local backdrop = {
		bgFile = lsm:Fetch("background", db.Frame.Background.Texture);
		edgeFile = lsm:Fetch("border", db.Frame.Border.Texture);
		tile = db.Frame.Background.Tile,
		tileSize = db.Frame.Background.TileSize,
		edgeSize = db.Frame.Border.EdgeSize,
		insets = {
			left = inset,
			right = inset,
			top = inset,
			bottom = inset,
		},
	};
	local bgColor = db.Frame.Background.Color;
	local borderColor = db.Frame.Border.Color;
	
	self.Frame:SetBackdrop(backdrop);
	self.Frame:SetBackdropColor(bgColor.r, bgColor.g, bgColor.b, bgColor.a);
	self.Frame:SetBackdropBorderColor(borderColor.r, borderColor.g, borderColor.b, borderColor.a);
	
	self.Frame.Container:ClearAllPoints();
	self.Frame.Container:SetPoint("TOP", self.Frame.Drag, "BOTTOM", 0, -5);
	self.Frame.Container:SetPoint("BOTTOM", self.Frame, "BOTTOM", 0, inset);
	self.Frame.Container:SetPoint("LEFT", self.Frame, "LEFT", inset, 0);
	self.Frame.Container:SetPoint("RIGHT", self.Frame, "RIGHT", -inset, 0);
	
	self:UpdateSize();
end

function addon:UpdateFrameFonts()
	local font = lsm:Fetch("font", db.Frame.Text.Font);
	local size = db.Frame.Text.Size;
	local flags = db.Frame.Text.Outline;
	local color = db.Frame.Text.Color;
	for i, v in ipairs(self.TextFrames) do
		v.Text:SetFont(font, size, flags);
		v.Text:SetTextColor(color.r, color.g, color.b, color.a);
	end
end

function _TableDump(arr, level)
	local text, value, padding, item = "", nil, "", nil;
	if level == nil then
		level = 0;
	end
	if level > 0 then
		for i = 0, level do
			padding = padding .. "    ";
		end
	end
	if type(arr) == "table" then
		for item, value in pairs(arr) do
			if type(item) == 'string' then
				item = '"' .. item .. '"';
			elseif type(item) == 'number' then
				item = tostring(item);
			end
			if type(value) == 'table' then
				text = text .. padding .. '[' .. tostring(item) .. '] = {\n' .. _TableDump(value, level + 1) .. padding .. "},\n";
			elseif type(value) == 'string' then
				text = text .. padding .. '[' .. tostring(item) .. '] = "' .. value .. '",\n';
			else
				text = text .. padding .. '[' .. tostring(item) .. '] = ' .. (value and tostring(value) or 'nil') .. ',\n';
			end
		end
	end
	return text;
end

function addon:TableDump(arr)
	local str = _TableDump(arr, 0);
	for v in str:gmatch("[^\n]+") do
		print(v);
	end
end

--[[
Output will appear here
["lfgMode"] = { -- queued
        [1] = "queued",
        [2] = "empowered",
},
["lfgMode"] = { -- ready check
        [1] = "proposal",
        [2] = "accepted",
},
["lfgMode"] = { -- in party
        [1] = "lfgparty",
},
["lfgMode"] = { -- after dungeon clear, still in party
        [1] = "lfgparty",
},

Output will appear here
["GetLFGMode"] = { -- not in a queue
},
["GetLFGQueueStats"] = {
},
["GetLFGMode"] = { -- queued, waiting on data
        [1] = "queued",
        [2] = "empowered",
},
["GetLFGQueueStats"] = {
},
["GetLFGMode"] = { -- queued, data ready
        [1] = "queued",
        [2] = "empowered",
},
["GetLFGQueueStats"] = {
        [1] = true,
        [2] = 0,
        [3] = 1,
        [4] = 1,
        [5] = 0,
        [6] = 6,
        [7] = "Random Lich King Dungeon",
        [8] = 582,
        [9] = 5,
        [10] = -1,
        [11] = 667,
        [12] = 667,
},
["GetLFGMode"] = { -- in a dungeon, not in a party
	[1] = "abandonedInDungeon",
},
["GetLFGQueueStats"] = {
},
--]]

--[[
["point"] = {
	[1] = "TOPLEFT",
	[2] = {
		[0] = userdata: 18848164,
		["Damage3"] = {
			[0] = userdata: 287CE440,
			["cover"] = {
				[0] = userdata: 0E52C5F8,
			},
			["texture"] = {
				[0] = userdata: 0E52D27C,
			},
		},
		["RoleIcon1"] = {
			[0] = userdata: 0E52B548,
		},
		["Damage2"] = {
			[0] = userdata: 2A67565C,
			["cover"] = {
				[0] = userdata: 0E52C494,
			},
			["texture"] = {
				[0] = userdata: 0E52C330,
			},
		},
		["LookingFor"] = {
			[0] = userdata: 0E8BEF38,
		},
		["Damage1"] = {
			[0] = userdata: 286ECE70,
			["cover"] = {
				[0] = userdata: 0E52C1CC,
			},
			["texture"] = {
				[0] = userdata: 0E52C068,
			},
		},
		["Title"] = {
			[0] = userdata: 0E8BECB0,
		},
		["Healer1"] = {
			[0] = userdata: 19085C6C,
			["cover"] = {
				[0] = userdata: 0E52BF04,
			},
			["texture"] = {
				[0] = userdata: 0E52BC3C,
			},
		},
		["Tank1"] = {
			[0] = userdata: 1AC09B68,
			["cover"] = {
				[0] = userdata: 0E52BAD8,
			},
			["texture"] = {
				[0] = userdata: 0E52B974,
			},
		},
		["Statistic"] = {
			[0] = userdata: 0E8BEDF4,
		},
		["RoleIcon2"] = {
			[0] = userdata: 0E52B6AC,
		},
		["RoleIcon3"] = {
			[0] = userdata: 0E52B810,
		},
	},
	[3] = "TOPLEFT",
	[4] = 0,
	[5] = 0,
},

["point"] = {
	[1] = "LEFT",
	[2] = {
		[0] = userdata: 0E8BEF38,
	},
	[3] = "RIGHT",
	[4] = 2.0000001405773,
	[5] = 0,
},

["point"] = {
	[1] = "BOTTOMRIGHT",
	[2] = {
		[0] = userdata: 18848164,
		["Damage3"] = {
			[0] = userdata: 287CE440,
			["cover"] = {
				[0] = userdata: 0E52C5F8,
			},
			["texture"] = {
				[0] = userdata: 0E52D27C,
			},
		},
		["RoleIcon1"] = {
			[0] = userdata: 0E52B548,
		},
		["Damage2"] = {
			[0] = userdata: 2A67565C,
			["cover"] = {
				[0] = userdata: 0E52C494,
			},
			["texture"] = {
				[0] = userdata: 0E52C330,
			},
		},
		["LookingFor"] = {
			[0] = userdata: 0E8BEF38,
		},
		["Damage1"] = {
			[0] = userdata: 286ECE70,
			["cover"] = {
				[0] = userdata: 0E52C1CC,
			},
			["texture"] = {
				[0] = userdata: 0E52C068,
			},
		},
		["Title"] = {
			[0] = userdata: 0E8BECB0,
		},
		["Healer1"] = {
			[0] = userdata: 19085C6C,
			["cover"] = {
				[0] = userdata: 0E52BF04,
			},
			["texture"] = {
				[0] = userdata: 0E52BC3C,
			},
		},
		["Tank1"] = {
			[0] = userdata: 1AC09B68,
			["cover"] = {
				[0] = userdata: 0E52BAD8,
			},
			["texture"] = {
				[0] = userdata: 0E52B974,
			},
		},
		["Statistic"] = {
			[0] = userdata: 0E8BEDF4,
		},
		["RoleIcon2"] = {
			[0] = userdata: 0E52B6AC,
		},
		["RoleIcon3"] = {
			[0] = userdata: 0E52B810,
		},
	},
	[3] = "BOTTOMRIGHT",
	[4] = 0,
	[5] = 0,
},

--]]